import asyncio
import functools

# every coroutine


async def coromask(coro, args, kwargs, fargs):
    """
    A coroutine that mask another coroutine  callback with args, and a
    function callbacks who manage input/output of corotine callback

    :param coro: is a coroutine object defined by the developer
    :param args: the list of arguments to run on the corotine *coro*
    :param fargs: the function that process the input and create an output related with the coro result

    :returns: a result, is a list of the elements for future argument
    """
    try:
        _in = args
        msg = ("Coromask args %s, kwargs %s, in coro %s" %
               (args, kwargs, coro))
        obtained = await coro(*args, **kwargs)
        if isinstance(obtained, Exception):
            raise Exception()
        else:
            result = fargs(_in, obtained)
            return result
    except Exception:
        raise Exception(msg)


def renew(task, coro, fargs, *args, **kwargs):
    """
    A simple function who manages the scheduled task and set the
    renew of the task

    :param task: is a Future initialized coroutine but not executed yet
    :param coro: is the corutine to renew when the first is finished
    :param fargs: the function to process input/output
    :param args: the unpacked list of extra arguments
    """
    if not task.cancelled():
        try:
            result = task.result()
            result_args, result_kwargs = result
            loop = asyncio.get_event_loop()
            task = loop.create_task(
                coromask(coro, result_args, result_kwargs, fargs))
            task.add_done_callback(functools.partial(renew, task, coro, fargs))
        except asyncio.InvalidStateError as ie:
            print("Invalid State Error", ie,
                  "Coro", coro,
                  "args", args,
                  "kargs", kwargs)
            raise ie
        except Exception as e:
            print("Resultado", result, task, coro, fargs)
            print("Excepcion en renew", e)
            raise e
    else:
        try:
            loop = task.get_loop()
            loop.stop()
            return False
        except Exception as e:
            raise e


def simple_fargs(_in, obtained):
    """
    Simple function who can be used in callback on coromask, the
    inputs are /_in/ and /obtained/ value from the coroutine executed.
    Return _in

    :_in: the input list
    :param obtained: the object that came from the result of coroutine execution

    :returns: _in
    """
    return _in


def simple_fargs_out(_in, obtained):
    """
    Simple function who can be used in callback on coromask, the
    inputs are /_in/ and /obtained/ value from the coroutine executed.
    Return obtained

    :param _in: the input list
    :param obtained: the object that came from the result of coroutine execution

    :returns: obtained
    """
    return obtained



#http://stackoverflow.com/questions/24687061/can-i-somehow-share-an-asynchronous-queue-with-a-subprocess

from multiprocessing import Manager, cpu_count
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import asyncio

def AsyncProcessQueue(maxsize=0):
    m = Manager()
    q = m.Queue(maxsize=maxsize)
    return _ProcQueue(q)


class _ProcQueue(object):
    def __init__(self,q):
        self._queue = q
        self._real_executor = None
        self._cancelled_join = False

    @property
    def _executor(self):
        if not self._real_executor:
            self._real_executor = ThreadPoolExecutor(max_workers=cpu_count())
        return self._real_executor

    def __getstate__(self):
        self_dict = self.__dict__
        self_dict['_real_executor'] = None
        return self_dict

    def __getattr__(self, name):
        if name in ['qsize', 'empty', 'full', 'put', 'put_nowait',
                    'get', 'get_nowait', 'close']:
            return getattr(self._queue, name)
        else:
            raise AttributeError("'%s' object has no attribute '%s'" % 
                                    (self.__class__.__name__, name))

    async def coro_put(self, item):
        loop = asyncio.get_event_loop()
        return (await loop.run_in_executor(self._executor, self.put, item))

    async def coro_get(self):
        loop = asyncio.get_event_loop()
        return (await loop.run_in_executor(self._executor, self.get))

    def cancel_join_thread(self):
        self._cancelled_join = True
        self._queue.cancel_join_thread()

    def join_thread(self):
        self._queue.join_thread()
        if self._real_executor and not self._cancelled_join:
            self._real_executor.shutdown()

async def _do_coro_proc_work(q, stuff, stuff2):
    ok = stuff + stuff2
    print("Passing %s to parent" % ok)
    await  q.coro_put(ok)  # Non-blocking
    item = q.get() # Can be used with the normal blocking API, too
    print("got %s back from parent" % item)

def do_coro_proc_work(q, stuff, stuff2):
    loop = asyncio.get_event_loop()
    loop.run_until_complete(_do_coro_proc_work(q, stuff, stuff2))

async def do_work(q):
    loop = asyncio.get_event_loop()
    await loop.run_in_executor(ProcessPoolExecutor(max_workers=1),do_coro_proc_work, q, 1, 2)
    item = await q.coro_get()
    print("Got %s from worker" % item)
    item = item + 25
    q.put(item)

if __name__  == "__main__":
    q = AsyncProcessQueue()
    loop = asyncio.get_event_loop()
    loop.run_until_complete(do_work(q))



# standar library
import asyncio
import functools
from asyncio import shield

# contrib
from networktools.library import my_random_string
from networktools.colorprint import gprint, bprint, rprint

# same module
from .taskloop import coromask, renew, simple_fargs, simple_fargs_out
from networktools.factory import FactoryClass


class TaskScheduler:
    """
    This class is a generic tasks scheduler that uses
    asyncio and multiprocessing
    For every new process activate M idle coro tasks
    until when are assigned by a function that send
    an id

    :param args: an unpacked list
    :param kwargs: an unpacked dictionary with at least {*ipt* dict, *ico* dict, *assigned_task* dict, *nproc* int, *sta_init* dict}, all in shared memory manager

    """

    def __init__(self, *args, **kwargs):
        """
        Is a generic init method, the inputs required are
        """
        self.ipt = kwargs.get('ipt', [])
        self.ico = kwargs.get('ico', [])
        self.assigned_tasks = kwargs.get('assigned_tasks', {})
        self.lnproc = kwargs.get('nproc', 3)
        self.sta_init = kwargs.get("sta_init", {})
        self.enqueued = kwargs.get('enqueued', list())

    def set_ipt(self, uin=4):
        """
        Defines a new id for relation process-collect_task, check if exists

        :param uin: is an optional int to define the length of the keys for ipt dict

        :returns: a new key from ipt
        """
        ipt = my_random_string(uin)
        while True:
            if ipt not in self.ipt:
                self.ipt.append(ipt)
                break
            else:
                ipt = my_random_string(uin)
        return ipt

    def set_ico(self, uin=4):
        """
        Defines a new id for task related to collect data
        inside a worker, check if exists

        :param uin: is an optional int to define the length of the keys for ico dict

        :returns: a new key from ico
        """
        ico = my_random_string(uin)
        while True:
            if ico not in self.ico:
                self.ico.append(ico)
                break
            else:
                ico = my_random_string(uin)
        return ico

    async def run_task(self, *args):
        """
        A default coroutine that await .1 secs every time

        :param args: a unpacked list

        :returns: the same list
        """
        await asyncio.sleep(.1)
        return args

    def set_new_run_task(self, **coros_callback):
        """
        Is a function that you have to call to define the corutines which you need
        to execute on the wheel (gear)

        Is recomended the use at the end of the __init__ method on your new class.

        :param coros_callback: is a dictionary with the corutines, that depends of what you need:

        - A unidirectional machine: you have to set only the *run_task* key
        - A bidirectional machine: you ahve to set also *net2service* and *service2net* keys
        """
        run_task = coros_callback.get('run_task', None)
        net2service_task = coros_callback.get('net2service', {})
        service2net_task = coros_callback.get('service2net', {})
        self.n2s = None
        self.s2n = None
        self.run_task = None

        if asyncio.iscoroutinefunction(run_task):
            self.run_task = run_task
        else:
            print("The callback run_task is not a coroutine")

        if asyncio.iscoroutinefunction(net2service_task.get('coro')):
            self.n2s = net2service_task
        else:
            print("The callback net2service is not a coroutine")

        if asyncio.iscoroutinefunction(service2net_task.get('coro')):
            self.s2n = service2net_task
        else:
            print("The callback service2net is not a coroutine")

    async def process_sta_task(self, ipt, ico, *args, **kwargs):
        """
        Process gather data for ids station

        This coroutine is executed for both cases: unidirectional and bidirectional

        :param ipt: is an ipt value to work with the *ipt* process
        :param ico: is an ico value to work with the *ico* object
        :param args: an unpacked list to execute the *run_task*

        :returns: the first two inputs and the unpacked result
        """
        assigned_task = self.assigned_tasks.get(ipt, {}).get(ico, None)
        result_args, result_kwargs = [], {}
        log = kwargs.get('log')
        try:
            if assigned_task:
                if not self.sta_init.get(assigned_task):
                    args, kwargs = self.set_pst(assigned_task, args, kwargs)
                result_args, result_kwargs = await shield(self.run_task(*args, **kwargs))
            else:
                result_args, result_kwargs = args, kwargs
            return [ipt, ico, *result_args], result_kwargs
        except Exception as e:
            if log:
                log.exception(
                    "STA_TASK: Falla en inicializar la corutina loop, %s %s" % (args, kwargs))
            raise e

    async def process_sta_task_n2s(self, ipt, ico, *args, **kwargs):
        """
        Process channel network to service data for ids station

        This coroutine is executed for bidirectional case, on network to service direction

        :param ipt: is an ipt value to work with the *ipt* process
        :param ico: is an ico value to work with the *ico* object
        :param args: an unpacked list to execute the *run_task*

        :returns: the first two inputs and the unpacked result
        """

        assigned_task = self.assigned_tasks.get(ipt, {}).get(ico, None)
        result_args, result_kwargs = [], {}
        log = kwargs.get('log')
        try:
            if assigned_task:
                if not self.sta_init.get(assigned_task):
                    args, kwargs = self.set_pst_n2s(
                        assigned_task, args, kwargs)
                result_args, result_kwargs = await shield(self.n2s_coro(*args, **kwargs))
            else:
                result_args, result_kwargs = args, kwargs
            return [ipt, ico, *result_args], result_kwargs
        except Exception as e:
            if log:
                log.exception(
                    "N2S: Falla en inicializar la corutina loop, %s %s" % (args, kwargs))
            raise e

    async def process_sta_task_s2n(self, ipt, ico, *args, **kwargs):
        """
        Process channel service 2 network for ids station

        This coroutine is executed for bidirectional case, on service to network direction

        :param ipt: is an ipt value to work with the *ipt* process
        :param ico: is an ico value to work with the *ico* object
        :param args: an unpacked list to execute the *run_task*

        :returns: the first two inputs and the unpacked result
        """
        assigned_task = self.assigned_tasks.get(ipt, {}).get(ico, None)
        result_args, result_kwargs = [], {}
        log = kwargs.get('log')
        try:
            if assigned_task:
                if not self.sta_init[assigned_task]:
                    args, kwargs = self.set_pst_s2n(
                        assigned_task, args, kwargs)
                result_args, result_kwargs = await shield(self.s2n_coro(*args, **kwargs))
            else:
                result_args, result_kwargs = args, kwargs
            return [ipt, ico, *result_args], result_kwargs
        except Exception as e:
            if log:
                log.exception(
                    "S2N: Falla en inicializar la corutina loop, %s %s" % (args, kwargs))
            raise e

    def get_ipt(self, ids):
        for ipt, ico_set in self.assigned_tasks.items():
            icos = [ico for ico in ico_set if ico_set[ico] == ids]
            rprint(icos)
            if icos:
                return ipt
        return None

    async def process_sta_manager(self, ipt):
        """
        Manage asignation of station to task inside ipt process

        :param ipt: the key of the process

        :returns: a list object with ipt value
        """
        #
        ids_list = self.proc_tasks[ipt]
        for ids in ids_list:
            if ids not in self.assigned_tasks[ipt].values():
                for ico, value in self.assigned_tasks[ipt].items():
                    if value is None:
                        self.assigned_tasks[ipt].update({ico: ids})
                        break
        await asyncio.sleep(1)
        return [ipt], {}

    def set_pst(self, ids, args, kwargs):
        """
        Set the factory for the wheel's array

        In your class you have to rewrite this.

        :param ids: key of the source
        :param args: list of arguments

        :returns: a different list of future arguments
        """
        return [ids, args[1], args[2]], kwargs

    def set_init(self, ids):
        """
        Is a flag for every station, the ids is related with the id
        assigned to the station

        """
        self.sta_init.update({ids: False})

    def set_init_args_kwargs(self, ipt):
        """
        Set the initial list of arguments

        In your class you have to rewrite this.

        :returns: a list of initial arguments
        """
        return [None, None, None, None], {}

    def add_sta_assigned(self, ipt, ico, ids):
        ico_d = self.assigned_tasks[ipt]
        ico_d[ico] = ids
        self.assigned_tasks[ipt] = ico_d

    def unset_sta_assigned(self, ipt, ico, ids):
        ico_d = self.assigned_tasks[ipt]
        ico_d[ico] = None
        self.assigned_tasks[ipt] = ico_d

    def add_task(self, ids, ipt):
        """
        Add an *ids* task to some *ipt* process

        :param ids: the key of a source
        :param ipt: the key or identifier of a process
        """
        self.proc_tasks[ipt] += [ids]
        icos = self.assigned_tasks[ipt]
        return icos

    def manage_tasks(self, ipt):
        """
        A method to manage the tasks assigned to *ipt* process

        Initialize an event loop, and assign idle tasks for this process

        Create the tasks for every source assigned to this process.
        Check the cases unidirectional and bidirectional.

        :param ipt: the key or identifier of a process
        """
        # loop = asyncio.get_event_loop()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        v = 1
        tasks = []
        self.assigned_tasks[ipt] = {}
        new_dict = {}
        for i in range(self.lnproc):
            ico = self.set_ico()
            nd = {ico:  None}
            new_dict.update(nd)
        self.assigned_tasks[ipt] = new_dict
        tasks_on_this_ipt = self.assigned_tasks.get(ipt)
        stax, sta_kwargs = self.set_init_args_kwargs(ipt)
        for ico in tasks_on_this_ipt.keys():
            if self.run_task:
                try:
                    args = [ipt, ico, *stax]
                    task = loop.create_task(
                        coromask(
                            self.process_sta_task,
                            args, sta_kwargs,
                            simple_fargs_out)
                    )
                    task.add_done_callback(
                        functools.partial(renew,
                                          task,
                                          self.process_sta_task,
                                          simple_fargs_out)
                    )
                    tasks.append(task)
                except Exception as ex:
                    print(
                        "Error en collect_task, gather stations, process_sta(task) %s, error %s"
                        % (ipt, ex))
                    print(ex)
                    raise ex

            # add if exists the other tasks
            try:
                if self.n2s:
                    self.n2s_coro = self.n2s.get('coro')
                    n2s_args = self.n2s.get('args')
                    n2s_args.insert(0, ipt)
                    args = [ipt, ico, *n2s_args]
                    task = loop.create_task(
                        coromask(
                            self.process_sta_task_n2s,
                            args,
                            simple_fargs_out)
                    )
                    task.add_done_callback(
                        functools.partial(renew,
                                          task,
                                          self.process_sta_task_n2s,
                                          simple_fargs_out)
                    )
                    tasks.append(task)
            except Exception as ex:
                print(
                    "Error en collect_task, gather stations, net2service(task) %s, error %s"
                    % (ipt, ex))
                print(ex)
                raise ex

            try:
                if self.s2n:
                    self.s2n_coro = self.s2n.get('coro')
                    s2n_args = self.s2n.get('args')
                    s2n_args.insert(0, ipt)
                    args = [ipt, ico, *s2n_args]
                    task = loop.create_task(
                        coromask(
                            self.process_sta_task_s2n,
                            args,
                            simple_fargs_out)
                    )
                    task.add_done_callback(
                        functools.partial(renew,
                                          task,
                                          self.process_sta_task_s2n,
                                          simple_fargs_out)
                    )
                    tasks.append(task)
            except Exception as exe:
                print(
                    "Error en collect_task, gather stations, service2net(task) %s, error %s"
                    % (ipt, exe))
                print(exe)
                raise exe
        try:
            args = [ipt]
            kwargs = {}
            task = loop.create_task(
                coromask(
                    self.process_sta_manager,
                    args, kwargs,
                    simple_fargs_out)
            )
            task.add_done_callback(
                functools.partial(renew,
                                  task,
                                  self.process_sta_manager,
                                  simple_fargs_out)
            )
            tasks.append(task)
        except Exception as exe:
            print("Error en collect_task, manager %s, error %s" % (ipt, exe))
            print(exe)
            raise exe
        bprint("Loop is running yet? IPT->%s" % ipt)
        rprint(loop.is_running())
        if not loop.is_running():
            loop.run_forever()



from setuptools import setup, find_packages
from pathlib import Path

path = Path(__file__).resolve().parent
with open(path/'README.md', encoding='utf-8') as f:
    long_description = f.read()

with open(path/'VERSION') as version_file:
    version = version_file.read().strip()


setup(name='tasktools',
      version=version,
      description='Some useful tools for asycnio Tasks: async while, the Scheduler and Assignator classes',
      url='https://tasktools.readthedocs.io/en/latest/',
      author='David Pineda Osorio',
      author_email='dpineda@uchile.cl',
      license='GPLv3',
      install_requires=['networktools'],      
      package_dir={'tasktools': 'tasktools'},
      package_data={
          'tasktools': ['../doc', '../docs', '../requeriments.txt']},
      packages=find_packages(),
      include_package_data=True,
      long_description=long_description,
      long_description_content_type='text/markdown',
      zip_safe=False)



import asyncio
import functools

from networktools.colorprint import gprint, bprint, rprint

from .taskloop import coromask, renew, simple_fargs, simple_fargs_out
from .scheduler import TaskScheduler


def get_free_ico(icos):
    ico_list = [
        key for key, v in icos.items() if not v]
    ico = None
    if ico_list:
        ico = ico_list.pop()
    return ico


class TaskAssignator:
    """
    Manage the tasks assigned to a TaskScheduler instance

    :param scheduler: A TaskScheduler instance
    :param queue_tasks: a queue
    :param sta_assigned: a dict managed for multiprocessing
    :param dt_status: a string that select the kind of group {GROUP, ALL}
    :param dt_group: a list of the group selected
    :param args: some extra args
    :param kwargs: some extra keyword args

    """

    def __init__(self,
                 scheduler,
                 queue_tasks,
                 queue_answer,
                 sta_assigned,
                 dt_status,
                 dt_group,
                 enqueued,
                 locker,
                 *args, **kwargs):
        self.scheduler = scheduler
        self.queue_tasks = queue_tasks
        self.queue_ans = queue_answer
        self.sta_assigned = sta_assigned
        self.dt_status = dt_status
        self.dt_group = dt_group
        self.enqueued = enqueued
        self.ts = 10
        self.locker = locker
        self.code_filter = kwargs.get('code_filter', 'code')
        self.counter = 0
        if 'ts' in kwargs:
            self.ts = kwargs.get('ts', 3)

    async def new_process(self, queue_tasks, * args, **kwargs):
        """
        This coroutine activate a process with new station task
        Check every \ts\ seconds the queue_tasks if there are new
        stations or tasksto add.

        Works on a async wheel

        :param queue_tasks: a queue to put task ids

        """
        await asyncio.sleep(self.ts)
        scheduler = self.scheduler
        sta_assigned = []
        for ipt, val in self.sta_assigned.items():
            for ico, ids in val.items():
                sta_assigned.append(ids)
        dt_status = self.dt_status
        dt_group = self.dt_group
        msg_in = []
        try:
            tasks = []
            W = 0
            if not queue_tasks.empty():
                for i in range(queue_tasks.qsize()):
                    ids = queue_tasks.get()
                    scheduler.status_tasks[ids] = True
                    scheduler.sta_init[ids] = True
                    if ids in self.enqueued:
                        q = 0
                        for ipt in scheduler.proc_tasks.keys():
                            q += 1
                            if len(scheduler.proc_tasks.get(ipt)) < \
                               scheduler.lnproc and \
                               ids not in sta_assigned:
                                respuesta = {'station': ids,
                                             'core': ipt, }
                                if dt_status == 'GROUP':
                                    if scheduler.stations.get(ids).get(
                                            self.code_filter) in dt_group:
                                        icos = scheduler.add_task(ids, ipt)
                                        scheduler.set_init(ids)
                                        ico = get_free_ico(icos)
                                        self.locker.acquire()
                                        self.scheduler.add_sta_assigned(
                                            ipt, ico, ids)
                                        self.locker.release()
                                        ans = "TASK %s ADDED TO %s" % (
                                            ids, ipt)
                                        respuesta.update({'added': True})
                                elif dt_status == 'ALL':
                                    icos = scheduler.add_task(ids, ipt)
                                    ico = get_free_ico(icos)
                                    scheduler.set_init(ids)
                                    qr = {ico: ids}
                                    self.counter += 1
                                    ans = "TASK %s ADDED TO %s, de un total de:%d" % (
                                        ids, ipt, self.counter)
                                    sta_assigned.append(ids)
                                    self.locker.acquire()
                                    self.scheduler.add_sta_assigned(
                                        ipt, ico, ids)
                                    self.locker.release()
                                    respuesta.update({'added': True})
                                self.queue_ans.put(respuesta)
                    queue_tasks.task_done()
            else:
                pass
                # self.queue_ans.put({'added': False})
            return [queue_tasks, *args], kwargs
        except Exception as ex:
            bprint("Error en asignación de tareas a procesador: %s" % ex)
            raise ex

    def new_process_task(self):
        """
        This function allows the system to call the coroutine that add
        a new_process in an asynchronous loop *the wheel*
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            args = [self.queue_tasks]
            # bprint("New process task")
            task = loop.create_task(
                coromask(
                    self.new_process,
                    args, {},
                    simple_fargs_out),
                name="task_new_process"
            )
            #
            task.add_done_callback(
                functools.partial(
                    renew,
                    task,
                    self.new_process,
                    simple_fargs_out)
            )
        except Exception as ex:
            rprint("Error en levantar corrutina %s" % ex)
            raise ex
        if not loop.is_running():
            loop.run_forever()
